import io
import logging
import re
from dataclasses import dataclass
from typing import List, Optional

logger = logging.getLogger(__name__)


@dataclass
class YeadocSnippet:
    code: str
    lineno: int
    id: str
    syntax: Optional[str] = None


def extract_from_buffer(f: io.TextIOBase, max_num_lines: int = 10000) -> List[YeadocSnippet]:
    out: List[YeadocSnippet] = []
    previous_nonempty_line = None
    k = 1

    while True:
        line = f.readline()
        k += 1
        if not line:
            # EOF
            break

        if line.strip() == "":
            continue

        if line.lstrip()[:3] == "```":
            syntax = line.strip()[3:]
            num_leading_spaces = len(line) - len(line.lstrip())
            lineno = k - 1
            # read the block
            code_block = []
            while True:
                line = f.readline()
                k += 1
                if not line:
                    raise RuntimeError("Hit end-of-file prematurely. Syntax error?")
                if k > max_num_lines:
                    raise RuntimeError(f"File too large (> {max_num_lines} lines). Set max_num_lines.")
                # check if end of block
                if line.lstrip()[:3] == "```":
                    break
                # Cut (at most) num_leading_spaces leading spaces
                nls = min(num_leading_spaces, len(line) - len(line.lstrip()))
                line = line[nls:]
                code_block.append(line)

            if previous_nonempty_line is None:
                previous_nonempty_line = line
                continue

            # check for keywords
            m = re.match(  # type: ignore
                r"<!--[-\s]*yeadoc-test:(.*)-->",
                previous_nonempty_line.strip(),
            )
            if m is None:
                pass  # ignore test because it is not labeled
            else:
                id = m.group(1).strip("- ")
                out.append(YeadocSnippet("".join(code_block), lineno, id, syntax))
                continue

        previous_nonempty_line = line

    return out


def load_tests_from_docstring(docstring: str) -> List[YeadocSnippet]:
    return extract_from_buffer(io.StringIO(docstring))



"""Command line entrypoint."""

import argparse
import sys
from typing import Callable, List, Optional

from yea import __version__, context, runner


if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class CliArgs:
    """Typed command line arguments."""

    plugin_args: List[str]

    def __init__(self, args: argparse.Namespace):
        self.action: Literal["run", "list", "r", "l"] = args.action
        self.all: bool = args.all if hasattr(args, "all") else False
        self.debug: bool = args.debug
        self.docs_only: bool = args.docs_only
        self.dryrun: bool = args.dryrun
        self.func: Callable = args.func
        self.live: bool = args.live
        self.platform: Optional[str] = args.platform
        self.shard: Optional[str] = args.shard
        self.suite: Optional[str] = args.suite
        self.tests: List[str] = args.tests
        self.plugin_args = args.plugin_args or []


def cli_list(yc: "context.YeaContext") -> None:
    print("Tests:")
    yc._args.action = "list"
    tr = runner.TestRunner(yc=yc)
    tests = tr.get_tests()
    test_ids = [len(t.test_id) for t in tests if t.test_id is not None]
    if None in test_ids:
        raise ValueError("Test ids must be set.")
    tlen = max(test_ids) if test_ids else 0
    for t in tests:
        print("  {:<{}s}: {}".format(t.test_id, tlen, t.name))
    tr.clean()


def cli_run(yc: "context.YeaContext") -> None:
    yc._args.action = "run"
    tr = runner.TestRunner(yc=yc)
    tr.run()


def cli() -> None:
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(dest="action", title="action", description="Action to perform")
    parser.add_argument("--debug", action="store_true", help="Print out extra debug info")
    parser.add_argument("--docs-only", help="Read tests from docstrings only", action="store_true")
    parser.add_argument("--dryrun", action="store_true", help="Do not do anything")
    parser.add_argument("--live", action="store_true", help="Run against real server")
    parser.add_argument("--shard", help="Specify testing shard")
    parser.add_argument("--suite", help="Specify testing suite")
    parser.add_argument("--platform", help="Specify testing platform")
    parser.add_argument("-p", "--plugin-args", action="append", help="Add plugin args")
    parser.add_argument("--version", help="Print version and exit", action="store_true")

    parse_list = subparsers.add_parser("list", aliases=["l"])
    parse_list.set_defaults(func=cli_list)
    parse_list.add_argument("tests", nargs="*")

    parse_run = subparsers.add_parser("run", aliases=["r"])
    parse_run.add_argument("-a", "--all", action="store_true", help="Run all")
    parse_run.add_argument("tests", nargs="*")
    parse_run.set_defaults(func=cli_run)
    args = parser.parse_args()

    if args.version:
        print("Yea {}".format(__version__))
        sys.exit(0)

    if not args.action:
        parser.print_help()
        sys.exit(1)

    cli_args = CliArgs(args)
    yc = context.YeaContext(args=cli_args)
    args.func(yc)


if __name__ == "__main__":
    cli()



"""Base TestlibConfig classes."""

from typing import Dict, List

import jsonschema  # type: ignore

import yaml

from .schema import default_filler, validator


def schema_violations_from_proposed_config(config: Dict) -> List[str]:

    schema_violation_messages = []
    for error in validator.iter_errors(config):
        schema_violation_messages.append(f"{error.message}")

    return schema_violation_messages


class TestlibConfig(dict):
    def __init__(self, d: Dict):
        super(TestlibConfig, self).__init__(d)

        if not isinstance(d, TestlibConfig):
            # ensure the data conform to the schema
            schema_violation_msgs = schema_violations_from_proposed_config(d)

            if len(schema_violation_msgs) > 0:
                err_msg = "\n".join(schema_violation_msgs)
                raise jsonschema.ValidationError(err_msg)

        # fill defaults not specified by user
        default_filler.validate(d)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return str(yaml.safe_dump(dict(self)))



"""Yea test class."""

import configparser
import functools
import itertools
import json
import os
import pathlib
import subprocess
import sys
import time
from typing import Any, Dict, List, Mapping, Optional, Tuple, Union

import requests

from yea import context, testcfg, testspec


def run_command(
    cmd_list: List[str],
    timeout: int = 300,
    env: Mapping = os.environ,
) -> int:
    print("INFO: RUNNING=", cmd_list)
    p = subprocess.Popen(cmd_list, env=env)
    try:
        p.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        print("ERROR: TIMEOUT")
        p.kill()
        try:
            p.communicate(timeout=30)
        except subprocess.TimeoutExpired:
            print("ERROR: double timeout")
            sys.exit(1)
    print("INFO: exit=", p.returncode)
    return p.returncode


def download(url: str, fname: str) -> bool:
    err: bool = False
    print(f"INFO: grabbing {fname} from {url}")
    try:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(fname, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
    except requests.exceptions.HTTPError as e:
        print("ERROR: url download error", url, e)
        err = True
    return err


def get_config(config: Dict[str, Any], prefix: str) -> Dict[str, Any]:
    """
    Recursively parse a "flat" config with column-separated key name definitions
    into a nested dictionary given a prefix.

    Example:
        config = {
            ":rug:ties_room_together": True,
            ":wandb:mock_server:lol": True,
            ":wandb:mock_server:lmao": False,
            ":wandb:mock_server:resistance:object": "Borg",
            ":wandb:mock_server:resistance:is_futile": True,
            ":wandb:foo": "bar",
        }
        prefix = ":wandb:"
        get_config(config, prefix)
        # {
        #     "mock_server": {
        #         "lol": True,
        #         "lmao": False,
        #         "resistance": {
        #             "object": "Borg",
        #             "is_futile": True,
        #         },
        #     },
        #     "foo": "bar",
        # }

    """
    # recursively get config values
    def parse(key: str, value: Any) -> Dict[str, Any]:
        if ":" not in key:
            return {key: value}
        else:
            key, subkey = key.split(":", 1)
            return {key: parse(subkey, value)}

    # recursively merge configs
    def merge(source: Dict[str, Any], destination: Dict[str, Any]) -> Dict[str, Any]:
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, {})
                merge(value, node)
            else:
                destination[key] = value
        return destination

    prefixed_items = {k[len(prefix) :]: v for (k, v) in config.items() if k.startswith(prefix)}
    parsed_items = []
    for k, v in prefixed_items.items():
        parsed_items.append(parse(k, v))

    return functools.reduce(merge, parsed_items) if parsed_items else {}


class YeaTest:
    def __init__(self, *, tname: pathlib.Path, yc: "context.YeaContext") -> None:
        self._tname = tname
        self._yc = yc
        self._args = yc._args
        self._retcode: int
        self._test_cfg: testcfg.TestlibConfig
        self._covrc: Optional[pathlib.Path] = None
        self._time_start: Optional[Union[int, float]] = None
        self._time_end: Optional[Union[int, float]] = None
        self._permute_groups: Optional[List[Any]] = None
        self._permute_items: Optional[Tuple[Any, ...]] = None

    def __str__(self) -> str:
        return f"{self._tname}"

    def _depend_files(self) -> bool:
        err = False
        dep = self._test_cfg.get("depend", {})
        files = dep.get("files", [])
        for fdict in files:
            fn = fdict["file"]
            fs = fdict["source"]
            assert fs.startswith("http"), "only http schemes supported right now"
            err = err or download(fs, fn)
        return err

    def _depend_install(self) -> bool:
        err = False
        req_list = self._test_cfg.get("depend", {}).get("requirements", [])
        options = self._test_cfg.get("depend", {}).get("pip_install_options", [])
        timeout = self._test_cfg.get("depend", {}).get("pip_install_timeout")
        if not (req_list or options):
            return err
        if not options:
            options.append("-qq")
        for req in req_list:
            cmd_list = ["python", "-m", "pip", "install"]
            cmd_list.extend(options)
            if " " in req:
                cmd_list.extend(req.split(" "))
            else:
                cmd_list.append(req)
            exit_code = run_command(cmd_list, timeout=timeout)
            err = err or exit_code != 0
            if err:
                return err
        return False

    def _depend_uninstall(self) -> bool:
        err = False
        req = self._test_cfg.get("depend", {}).get("uninstall", [])
        timeout = self._test_cfg.get("depend", {}).get("pip_uninstall_timeout")
        if not req:
            return err
        fname = ".yea-uninstall.txt"
        with open(fname, "w") as f:
            f.writelines(f"{item}\n" for item in req)
        cmd_list = ["python", "-m", "pip", "uninstall", "-qq", "-y", "-r", fname]
        exit_code = run_command(cmd_list, timeout=timeout)
        if os.path.exists(fname):
            os.remove(fname)
        err = err or exit_code != 0
        return err

    def _depend(self) -> bool:
        tname = self._tname
        print("INFO: DEPEND=", tname)
        tpath = pathlib.Path(tname)
        os.chdir(tpath.parent)
        err = False
        err = err or self._depend_uninstall()
        err = err or self._depend_files()
        err = err or self._depend_install()

        return err

    def _run(self) -> None:
        tname = self._tname
        print("INFO: RUN=", tname)
        program = self._test_cfg.get("command", {}).get("program")
        # test execution mode: default (./module/lib.py) or module (python -m module.lib)
        mode = self._test_cfg.get("command", {}).get("mode")
        tpath = pathlib.Path(tname)
        os.chdir(tpath.parent)
        if program is None:
            cmd = [f"./{tpath.name}"]
        elif mode == "module":
            cmd = ["-m", program.split(".py")[0].replace("/", ".")]
        else:
            cmd = [f"./{program}"]
        cmd_list = ["coverage", "run"]
        if self._covrc:
            cmd_list.extend(["--rcfile", str(self._covrc)])
        cmd_list.extend(cmd)
        cmd_cfg = self._test_cfg.get("command", {})
        args = cmd_cfg.get("args", [])
        timeout = cmd_cfg.get("timeout")
        cmd_list.extend(args)
        env = os.environ.copy()
        elist = self._test_cfg.get("env", [])
        for edict in elist:
            env.update(edict)
        if self._permute_groups and self._permute_items:
            env["YEA_PARAM_NAMES"] = ",".join(self._permute_groups)
            env["YEA_PARAM_VALUES"] = ",".join(map(str, self._permute_items))

        plugins = self._test_cfg.get("plugin", [])
        params = (
            {k: v for (k, v) in zip(self._permute_groups, self._permute_items)}
            if self._permute_groups and self._permute_items
            else None
        )
        if plugins:
            for plugin_name in plugins:
                prefix = f":{plugin_name}:"

                if params is not None:
                    # need to configure the plugin?
                    # get the plugin by its name
                    plug = self._yc._plugs.get_plugin(plugin_name)
                    plugin_params = get_config(params, prefix)
                    if plugin_params:
                        plug.monitors_configure(plugin_params)

                # process vars
                penv = plugin_name.upper()
                pnames = []
                pvalues = []
                for items in self._test_cfg.get("var", []):
                    for k, v in items.items():
                        if k.startswith(prefix):
                            pnames.append(k[len(prefix) :])
                            pvalues.append(v)
                if pnames and pvalues:
                    env[f"YEA_PLUGIN_{penv}_NAMES"] = ",".join(pnames)
                    env[f"YEA_PLUGIN_{penv}_VALUES"] = json.dumps(pvalues)
            env["YEA_PLUGINS"] = ",".join(plugins)

        exit_code = run_command(cmd_list, env=env, timeout=timeout)
        self._retcode = exit_code

    def _load(self) -> None:
        spec = None
        # load yea file if exists
        fname = str(self._tname)
        if fname.endswith(".py"):
            yea_name = fname[:-3] + ".yea"
            if os.path.exists(yea_name):
                spec = testspec.load_yaml_from_file(yea_name)
        elif fname.endswith(".yea"):
            spec = testspec.load_yaml_from_file(fname)
        if not spec:
            docstr = testspec.load_docstring(self._tname)
            spec = testspec.load_yaml_from_docstring(docstr)
        # print("SPEC:", self._tname, spec)
        cfg = testcfg.TestlibConfig(spec)
        # print("TESTCFG", cfg)
        self._test_cfg = cfg

    def _prep(self) -> None:
        """Cleanup and/or populate wandb dir."""
        self._yc.test_prep(self)
        # load file and docstring eval criteria
        self._setup_coverage_file()
        self._setup_coverage_config()

    def _setup_coverage_file(self) -> None:
        # dont mess with coverage_file (for now) if already set
        if self._yc._covfile is not None:
            return
        covfname = ".coverage-{}-{}".format(self._yc._pid, self.test_id)
        covfile = self._yc._cachedir.joinpath(covfname)
        os.environ["COVERAGE_FILE"] = str(covfile)

    def _setup_coverage_config(self) -> None:
        # do we have a template?
        template = self._yc._cfg._coverage_config_template
        if not template:
            return
        cov_src = self._yc._cfg._coverage_source
        cov_env = self._yc._cfg._coverage_source_env
        if cov_env is not None:
            cov_env_src = os.environ.get(cov_env)
            if cov_env_src:
                cov_src = cov_env_src
        if cov_src is None:
            return

        # find our sources
        #   source is from tox (set in env)
        #   or set from conf (if not in env)
        if self._yc._cfg._cfroot is None:
            raise RuntimeError("_cf_root not set")
        p = self._yc._cfg._cfroot.joinpath(template)
        if not p.exists():
            raise RuntimeError(f"Coverage config template {p} does not exist")

        cf = configparser.ConfigParser()
        cf.read(p)

        cf["run"]["source"] = cov_src

        covrc_fname = f"yea-covrc-{self._yc._pid}-{self.test_id}.conf"
        covrc = self._yc._cachedir.joinpath(covrc_fname)
        with open(covrc, "w") as configfile:
            cf.write(configfile)

        self._covrc = covrc

    def _fin(self) -> None:
        """Reap anything in wandb dir"""
        self._yc.test_done(self)

    def run(self) -> None:
        self._prep()
        if not self._args.dryrun:
            err = self._depend()
            # TODO: record error instead of assert
            assert not err, "Problem getting test dependencies"
            self._time_start = time.time()
            self._run()
            self._time_end = time.time()
        self._fin()

    def get_permutations(self) -> List["YeaTest"]:
        self._load()
        params = self._test_cfg.get("parametrize")
        if not params:
            return [self]
        groups = params.get("permute", [])
        gnames = []
        glist = []
        for g in groups:
            assert isinstance(g, dict)
            assert len(g) == 1
            k = next(iter(g))
            gnames.append(k)
            v = g[k]
            assert isinstance(v, list)
            glist.append(v)
        items = list(itertools.product(*glist))
        r = []
        for tnum, it in enumerate(items):
            t = YeaTest(tname=self._tname, yc=self._yc)
            t._load()
            t._permute_groups = gnames
            t._permute_items = it
            tpname = f"{tnum}-{'-'.join(map(str, it))}"
            tid = t._test_cfg.get("id")
            if tid:
                t._test_cfg["id"] = f"{tid}.{tpname}"
            r.append(t)
        return r

    @property
    def name(self) -> str:
        root = self._yc._cfg._cfroot
        if root is None:
            raise TypeError("Config root not set")
        b = self._tname.relative_to(root)
        return str(b)

    @property
    def test_id(self) -> Optional[str]:
        tid = str(self._test_cfg.get("id", "")) if self._test_cfg else None
        return tid

    @property
    def _sort_key(self) -> str:
        tid = str(self._test_cfg.get("id", "")) if self._test_cfg else ""
        return tid + ":" + self.name

    @property
    def config(self) -> testcfg.TestlibConfig:
        return self._test_cfg



"""Config class."""

import configparser
import os
import re
from pathlib import Path
from typing import List, Optional


class Config:
    def __init__(self) -> None:
        self._coverage_config_template: Optional[str] = None
        self._coverage_source: Optional[str] = None
        self._coverage_source_env: Optional[str] = None
        self._cfname: Optional[Path] = None
        self._cfroot: Optional[Path] = None
        self._cf = None
        self._test_dirs: List[str] = []
        found = self._find_config()
        if found:
            cf = self._load_config()
            self._parse_config(cf)
            self._cf = cf
        else:
            self._cfroot = Path(".")
            self._test_dirs = ["."]

    def _find_config(self) -> bool:
        p = Path.cwd()
        # TODO: change to use parents
        while True:
            cf = Path(p, ".yearc")
            if cf.is_file():
                self._cfname = cf
                self._cfroot = p
                return True
            gd = Path(p, ".git")
            if gd.is_dir():
                return False
            n = p.parent
            if n.samefile(p):
                break
            p = n
        return False

    def _load_config(self) -> configparser.ConfigParser:
        p = self._cfname
        if p is None:
            raise ValueError("No config file found")
        cf = configparser.ConfigParser()
        cf.read(p)
        return cf

    def _parse_config(self, cf: configparser.ConfigParser) -> List[str]:
        ycfg = cf.items("yea")
        ydict = dict(ycfg)
        test_paths = ydict.get("test_paths", "")
        test_list = re.findall(r"[\S]+", test_paths)
        self._test_dirs = test_list
        # test_paths = ydict.get("test_paths", "")

        self._coverage_config_template = ydict.get("coverage_config_template", "")
        self._coverage_source = ydict.get("coverage_source")
        # TODO: clean up how this works, user could already have an absolute path
        if self._coverage_source and self.test_root:
            self._coverage_source = os.path.join(self.test_root, self._coverage_source)
        self._coverage_source_env = ydict.get("coverage_source_env")
        self._results_file = ydict.get("results_file")

        return test_list

    @property
    def test_dirs(self) -> List[str]:
        return self._test_dirs

    @property
    def test_root(self) -> Optional[Path]:
        return self._cfroot



# -*- coding: utf-8 -*-
"""test runner."""
import ast
import logging
import os
import pathlib
import re
import shutil
import sys
from typing import Any, Dict, Generator, List, Optional, Union

import junit_xml  # type: ignore

from yea import context, testspec, ytest

from .yeadoc import YeadocSnippet, load_tests_from_docstring


logger = logging.getLogger(__name__)


def convert(text: str) -> Union[int, str]:
    return int(text) if text.isdigit() else text.lower()


def alphanum_sort(key: "ytest.YeaTest") -> List[Union[int, str]]:
    return [convert(c) for c in re.split("([0-9]+)", key._sort_key)]


class TestRunner:
    def __init__(self, *, yc: "context.YeaContext"):
        self._yc = yc
        self._cfg = yc._cfg
        if self._cfg.test_root is None:
            raise TypeError("test_root is not set")
        self._tmpdir = pathlib.Path(self._cfg.test_root, ".yeadoc")
        self.prepare()
        self._args = yc._args
        self._test_files: List[str] = []
        self._results: List[junit_xml.TestCase] = []
        self._test_list: List["ytest.YeaTest"] = []
        self._populate()

    def prepare(self) -> None:
        if self._tmpdir.exists():
            shutil.rmtree(self._tmpdir)
        self._tmpdir.mkdir()

    def clean(self) -> None:
        if self._tmpdir.exists():
            shutil.rmtree(self._tmpdir)

    def _get_args_list(self) -> Optional[List[str]]:
        # TODO: clean up args parsing
        args_tests = getattr(self._args, "tests", None)
        if not args_tests:
            return None

        alist = []
        for a in args_tests:
            p = pathlib.Path(a)
            p = p.resolve()
            alist.append(str(p))
        return alist

    def _get_dirs(self) -> Generator[pathlib.Path, None, None]:
        # generate to return all test dirs and recursively found dirs
        if self._cfg.test_root is None:
            raise TypeError("test_root is not set")
        for tdir in self._cfg.test_dirs:
            path_dir = pathlib.Path(self._cfg.test_root, tdir)
            yield path_dir
            for root, dirs, _ in os.walk(path_dir, topdown=True):
                # TODO: temporary hack to avoid walking into wandb dir
                # use .gitignore instead
                if "wandb" in dirs:
                    dirs.remove("wandb")
                if ".tox" in dirs:
                    dirs.remove(".tox")
                for d in dirs:
                    path_dir = pathlib.Path(self._cfg.test_root, root, d)
                    yield path_dir

    def _get_platform(self) -> str:
        if self._args.platform:
            return self._args.platform
        p = sys.platform
        if p.startswith("win"):
            p = "win"
        elif p == "darwin":
            p = "mac"
        return p

    def _should_skip_test(self, spec: Any) -> bool:
        my_platform = self._get_platform()
        suite = spec.get("tag", {}).get("suite", "main")
        shards = spec.get("tag", {}).get("shards", [])
        shard = spec.get("tag", {}).get("shard", "default")
        platforms = spec.get("tag", {}).get("platforms", [])
        shards.append(shard)
        skip_all = spec.get("tag", {}).get("skip", False)
        skips = spec.get("tag", {}).get("skips", [])
        if skip_all:
            return True
        for skip in skips:
            skip_platform = skip.get("platform")
            # right now the only specific skip is platform, if not specified skip all
            if skip_platform is None:
                return True
            if skip_platform and my_platform == skip_platform:
                return True
        if self._args.suite and self._args.suite != suite:
            return True
        if self._args.shard and self._args.shard not in shards:
            return True
        if platforms and my_platform not in platforms:
            return True
        # if we specify platform, skip any platform that doesnt match
        if self._args.platform and my_platform not in platforms:
            return True
        return False

    def _populate(self) -> None:
        tpaths = []

        args_tests = self._get_args_list() or []

        all_tests = False
        if self._args.action == "list" and not self._args.tests:
            all_tests = True
        if self._args.action == "run" and self._args.all:
            all_tests = True

        if not self._args.docs_only:
            for path_dir in self._get_dirs():
                # TODO: look for .yea, or look for .py with docstring
                for tpath in path_dir.glob("*.py"):
                    if not all_tests and str(tpath) not in args_tests:
                        logger.debug("skip fname {}".format(tpath))
                        continue
                    docstr = testspec.load_docstring(tpath)
                    spec = testspec.load_yaml_from_docstring(docstr)
                    if not spec:
                        logger.debug("skip nospec {}".format(tpath))
                        continue

                    if all_tests:
                        if self._should_skip_test(spec):
                            continue

                    tpaths.append(tpath)
                for tpath in path_dir.glob("*.yea"):
                    # TODO: parse yea file looking for path info
                    spec = testspec.load_yaml_from_file(tpath)

                    # if program is specified, keep track of yea file
                    py_fname = spec.get("command", {}).get("program")
                    if py_fname:
                        # hydrate to full path, take base from tpath
                        py_fname = os.path.join(tpath.parent, py_fname)
                        t_fname = tpath
                    else:
                        py_fname = str(tpath)[:-4] + ".py"
                        t_fname = py_fname

                    if not os.path.exists(py_fname):
                        continue

                    if all_tests:
                        if self._should_skip_test(spec):
                            continue

                    if not all_tests:
                        if py_fname not in args_tests and str(tpath) not in args_tests:
                            logger.debug("skip yea fname {}".format(tpath))
                            continue

                    # add .yea or .py file
                    tpaths.append(pathlib.Path(t_fname))

        # pick up yea tests from docstrings
        id_test_map: Dict[str, YeadocSnippet] = {}
        for path_dir in self._get_dirs():
            # build up the list of tests that can be run by parsing docstrings
            for tpath in path_dir.glob("*.py"):

                # parse the test file using ast
                with open(tpath, encoding="utf8") as f:
                    mod = ast.parse(f.read())

                function_definitions = [node for node in mod.body if isinstance(node, ast.FunctionDef)]
                for func in function_definitions:
                    docstr = ast.get_docstring(func) or ""
                    snippets = load_tests_from_docstring(docstr)
                    for s in snippets:
                        id_test_map[s.id] = s

                classes = [node for node in mod.body if isinstance(node, ast.ClassDef)]
                for class_ in classes:
                    methods = [node for node in class_.body if isinstance(node, ast.FunctionDef)]
                    for func in methods:
                        docstr = ast.get_docstring(func) or ""
                        snippets = load_tests_from_docstring(docstr)
                        for s in snippets:
                            id_test_map[s.id] = s
                    class_docstr = ast.get_docstring(class_) or ""
                    if class_docstr is not None:
                        snippets = load_tests_from_docstring(class_docstr)
                        for s in snippets:
                            id_test_map[s.id] = s

        for path_dir in self._get_dirs():
            for tpath in path_dir.glob("*.yea"):
                # TODO: parse yea file looking for path info
                spec = testspec.load_yaml_from_file(tpath)
                id_not_in_map = spec.get("id", None) not in id_test_map
                test_selected_to_run = all_tests or str(tpath) in args_tests

                if id_not_in_map or not test_selected_to_run:
                    logger.debug("skip yea fname: {}".format(tpath))
                    continue

                if all_tests:
                    if spec.get("tag", {}).get("skip", False):
                        continue
                    suite = spec.get("tag", {}).get("suite", "main")
                    shard = spec.get("tag", {}).get("shard", "default")
                    if self._args.suite and self._args.suite != suite:
                        continue
                    if self._args.shard and self._args.shard != shard:
                        continue

                # write test and spec to tempfiles
                shutil.copy(tpath, self._tmpdir)
                t_fname = self._tmpdir / tpath.name
                py_fname = str(t_fname)[:-4] + ".py"
                with open(py_fname, "w") as f:
                    f.write(id_test_map[spec["id"]].code)

                # add .yea or .py file
                tpaths.append(pathlib.Path(py_fname))

        tlist = []
        for tname in tpaths:
            t = ytest.YeaTest(tname=tname, yc=self._yc)
            test_perms = t.get_permutations()
            tlist.extend(test_perms)

        tlist.sort(key=alphanum_sort)
        self._test_list = tlist

    def _runall(self) -> None:
        for t in self._test_list:
            self._yc.monitors_reset()
            t.run()
            self._capture_result(t)

    def _check_dict(
        self,
        result: List[str],
        s: Any,
        expected: Optional[dict],
        actual: dict,
    ) -> None:
        if expected is None:
            return
        for k, v in actual.items():
            exp = expected.get(k)
            if exp != v:
                result.append("BAD_{}({}:{}!={})".format(s, k, exp, v))
        for k, v in expected.items():
            act = actual.get(k)
            if v != act:
                result.append("BAD_{}({}:{}!={})".format(s, k, v, act))

    def _capture_result(self, t: "ytest.YeaTest") -> None:
        test_cfg = t._test_cfg
        if not test_cfg:
            return
        if self._yc.is_live():
            # we are live
            return
        result = self._yc.test_check(t)
        # print("GOTRES", result)
        result_str = ",".join(result)
        # self._results[t._tname] = result_str
        if t._time_end is None or t._time_start is None:
            raise RuntimeError("Test not run")
        elapsed = t._time_end - t._time_start
        tc = junit_xml.TestCase(t.test_id, classname="yea_func", elapsed_sec=elapsed)
        if result_str:
            tc.add_failure_info(message=result_str)
        self._results.append(tc)

    def run(self) -> None:
        try:
            self._populate()
            # inform so we only start monitors needed
            self._yc.monitors_inform(self.get_tests())
            self._yc.monitors_init()
            self._yc.monitors_start()
            self._runall()
            self.finish()
        finally:
            self._yc.monitors_stop()

    def _save_results(self) -> None:
        res_fname = self._yc._cfg._results_file
        if not res_fname:
            return
        if self._yc._cfg._cfroot is None:
            raise RuntimeError("No cfroot set")
        p = self._yc._cfg._cfroot.joinpath(res_fname)
        ts = junit_xml.TestSuite("yea-func", self._results)
        # create testfile dir if it doesnt exist
        testdir = p.parent  # get the directory portion of path
        testdir.mkdir(parents=True, exist_ok=True)
        with open(p, "w") as f:
            junit_xml.TestSuite.to_file(f, [ts], prettyprint=False, encoding="utf-8")

    def finish(self) -> None:
        self.clean()
        self._save_results()
        exit_code = 0
        print("\nResults:")
        print("--------")
        if not self._results:
            sys.exit(exit_code)
        tlen = max([len(tc.name) for tc in self._results])

        use_emoji = not sys.platform.startswith("win")
        for tc in self._results:
            # TODO: fix hack that only looks at first message
            r = tc.failures[0]["message"] if tc.failures else ""
            emoji = "😃" if use_emoji else ":)"
            if r:
                emoji = '🐴 "Neigh" -- ' if use_emoji else 'XX "Neigh" -- '

            print("  {:<{}s}: {}{}".format(tc.name, tlen, emoji, r))
            if r:
                exit_code = 1
        sys.exit(exit_code)

    def get_tests(self) -> List["ytest.YeaTest"]:
        return self._test_list



"""yea context."""

import datetime
import logging
import os
from pathlib import Path
from typing import Optional

from yea import cli, config, plugins, ytest


class YeaContext:
    def __init__(self, *, args: "cli.CliArgs"):
        self._args = args
        self._cachedir: Path
        self._covfile: Optional[str] = None
        self._now = datetime.datetime.now()
        self._ts = self._now.strftime("%Y%m%dT%H%M%S")
        self._pid = os.getpid()
        self._setup_env()
        self._cfg = config.Config()
        self._setup_cachedir()
        self._setup_logging()
        self._plugs: plugins.Plugins = plugins.Plugins(yc=self)

    def _setup_env(self) -> None:
        self._covfile = os.environ.get("COVERAGE_FILE")

    def _setup_cachedir(self) -> None:
        root = self._cfg._cfroot
        if root is None:
            raise RuntimeError("No config root.")
        p = root.joinpath(".yea_cache")
        if not p.exists():
            p.mkdir()
        self._cachedir = p

    def _setup_logging(self) -> None:
        logger = logging.getLogger("yea")
        logger.propagate = False
        logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter(fmt="%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
        logfname = "debug-{}-{}.log".format(self._ts, self._pid)
        lf = self._cachedir.joinpath(logfname)
        fh = logging.FileHandler(lf)
        fh.setFormatter(formatter)
        fh.setLevel(logging.DEBUG)
        logger.addHandler(fh)
        logger.info("started")

    def is_live(self) -> bool:
        return self._args.live

    def monitors_inform(self, tlist: list) -> None:
        self._plugs.monitors_inform(tlist)

    def monitors_init(self) -> None:
        self._plugs.monitors_init()

    def monitors_start(self) -> None:
        self._plugs.monitors_start()

    def monitors_stop(self) -> None:
        self._plugs.monitors_stop()

    def monitors_reset(self) -> None:
        self._plugs.monitors_reset()

    def test_prep(self, yt: "ytest.YeaTest") -> None:
        # wandb_dir_safe_cleanup()
        self._plugs.test_prep(yt)

    def test_done(self, yt: "ytest.YeaTest") -> None:
        # wandb_dir_safe_cleanup()
        self._plugs.test_done(yt)

    def test_check(self, yt: "ytest.YeaTest") -> list:
        # ctx = self._backend.get_state()
        result_list = self._plugs.test_check(yt)
        return result_list



"""
Utilities for parsing a test specification.
"""

import ast
import pathlib
from typing import Any, Dict, Optional, Union

import yaml


def load_docstring(filepath: Union[str, pathlib.Path]) -> str:
    with open(filepath, encoding="utf8") as fd:
        file_contents = fd.read()
    module = ast.parse(file_contents)
    docstring = ast.get_docstring(module)
    return docstring or ""


def find_yaml_str(s: str) -> Optional[str]:
    """Loads YAML from docstring."""
    split_lines = s.split("\n")

    # Cut YAML from rest of docstring
    for index, line in enumerate(split_lines):
        line = line.strip()
        if line == "---" or line.startswith("--- "):
            # TODO: validate, capture type
            # !<tag:wandb.ai,2021:yea>
            cut_from = index + 1
            yaml_string = "\n".join(split_lines[cut_from:])
            return yaml_string
    return None


def load_yaml_from_docstring(docstring: str) -> Dict[str, Any]:
    found = find_yaml_str(docstring)
    if not found:
        return dict()
    return load_yaml_from_str(found)


def load_yaml_from_str(yaml_string: str) -> Dict[str, Any]:
    try:
        data = dict(yaml.load(yaml_string, Loader=yaml.SafeLoader))
    except yaml.scanner.ScannerError:
        print("WARNING: Unable to parse yaml")
        data = {}
    return data


def load_yaml_from_file(filepath: Union[str, pathlib.Path]) -> Dict[str, Any]:
    file_contents = ""
    with open(filepath, encoding="utf8") as fd:
        file_contents = fd.read()
    found = find_yaml_str(file_contents)
    found = found or file_contents
    return load_yaml_from_str(found)



# lifted from github.com/wandb/sweeps

import json
from pathlib import Path
from typing import Any

import jsonschema  # type:ignore
from jsonschema import Draft7Validator, validators

testlib_config_jsonschema_fname = Path(__file__).parent / "schema-yea.json"
with open(testlib_config_jsonschema_fname, "r") as f:
    testlib_config_jsonschema = json.load(f)


format_checker = jsonschema.FormatChecker()


@format_checker.checks("float")  # type: ignore
def float_checker(value: Any) -> bool:
    return isinstance(value, float)


@format_checker.checks("integer")  # type: ignore
def int_checker(value: Any) -> bool:
    return isinstance(value, int)


validator = Draft7Validator(schema=testlib_config_jsonschema, format_checker=format_checker)


def extend_with_default(validator_class):  # type: ignore
    # https://python-jsonschema.readthedocs.io/en/stable/faq/#why-doesn-t-my-schema-s-default-property-set-the-default-on-my-instance
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):  # type: ignore

        errored = False
        for error in validate_properties(
            validator,
            properties,
            instance,
            schema,
        ):
            errored = True
            yield error

        if not errored:
            for property, subschema in properties.items():
                if "default" in subschema:
                    instance.setdefault(property, subschema["default"])

    return validators.extend(
        validator_class,
        {"properties": set_defaults},
    )


DefaultFiller = extend_with_default(Draft7Validator)  # type: ignore
default_filler = DefaultFiller(schema=testlib_config_jsonschema, format_checker=format_checker)



# -*- coding: utf-8 -*-
"""yea setup."""

from setuptools import setup


setup(
    name="yea",
    version="0.7.15",
    description="Test harness breaking the sound barrier",
    packages=["yea"],
    install_requires=[
        "coverage",
        "importlib-metadata>=3.0.0",
        "jsonschema",
        "junit-xml",
        "PyYAML",
        "requests",
    ],
    package_dir={"": "src"},
    entry_points={
        "console_scripts": [
            "yea=yea.cli:cli",
        ]
    },
    include_package_data=True,
    package_data={"": ["yea/schema-*.json"]},
    zip_safe=False,
    license="MIT license",
    python_requires=">=3.6",
)



import importlib
import multiprocessing as mp
import os
from typing import Dict


def _setup_params() -> Dict[str, str]:
    env_names = os.environ.get("YEA_PARAM_NAMES")
    env_values = os.environ.get("YEA_PARAM_VALUES")
    if env_names is None or env_values is None:
        return dict()
    names = env_names.split(",")
    values = env_values.split(",")
    params = dict(zip(names, values))
    return params


def setup_mp(params: Dict[str, str]) -> None:
    start_method = params.get(":yea:start_method")
    if not start_method:
        return
    print(f"INFO: start_method= {start_method}")
    mp.set_start_method(start_method)
    # TODO: check mp setup?


def setup_plugins() -> None:
    env_plugins = os.environ.get("YEA_PLUGINS")
    if env_plugins is None:
        return
    plugins = env_plugins.split(",")
    for plug in plugins:
        mod_name = f"yea_{plug}"
        mod = importlib.import_module(mod_name)
        mod_setup = getattr(mod, "setup", None)
        if not mod_setup:
            continue
        mod_setup()


def setup() -> None:
    p = _setup_params()
    setup_mp(params=p)
    setup_plugins()
